@@ -63,13 +63,13 @@ void bch2_backpointer_to_text(struct printbuf *out, struct bch_fs *c, struct bke
 	if (ca)
 		prt_printf(out, "bucket=%llu:%llu:%u ", bucket.inode, bucket.offset, bucket_offset);
 	else
-		prt_printf(out, "sector=%llu:%llu ", bp.k->p.inode, bp.k->p.offset >> MAX_EXTENT_COMPRESS_RATIO_SHIFT);
+		prt_printf(out, "sector=%llu:%llu ", bp.k->p.inode, bp.k->p.offset >> c->sb.extent_bp_shift);
 
 	bch2_btree_id_level_to_text(out, bp.v->btree_id, bp.v->level);
 	prt_str(out, " data_type=");
 	bch2_prt_data_type(out, bp.v->data_type);
 	prt_printf(out, " suboffset=%u len=%u gen=%u pos=",
-		   (u32) bp.k->p.offset & ~(~0U << MAX_EXTENT_COMPRESS_RATIO_SHIFT),
+		   (u32) bp.k->p.offset & ~(~0U << c->sb.extent_bp_shift),
 		   bp.v->bucket_len,
 		   bp.v->bucket_gen);
 	bch2_bpos_to_text(out, bp.v->pos);
@@ -155,8 +155,7 @@ int bch2_bucket_backpointer_mod_nowritebuffer(struct btree_trans *trans,
 				bool insert)
 {
 	CLASS(btree_iter, bp_iter)(trans, BTREE_ID_backpointers, bp->k.p,
-				   BTREE_ITER_intent|
-				   BTREE_ITER_with_updates);
+				   BTREE_ITER_intent);
 	struct bkey_s_c k = bkey_try(bch2_btree_iter_peek_slot(&bp_iter));
 
 	if (insert
@@ -441,7 +440,11 @@ static int drop_dev_and_update(struct btree_trans *trans, enum btree_id btree,
 	return bch2_btree_insert_trans(trans, btree, n, 0);
 }
 
-static int check_extent_checksum(struct btree_trans *trans,
+/*
+ * returns 0 if we didn't find a bad checksum, and did no work
+ * returns 1 if we dropped bad replica
+ */
+static int kill_replica_if_checksum_bad(struct btree_trans *trans,
 				 enum btree_id btree, struct bkey_s_c extent,
 				 enum btree_id o_btree, struct bkey_s_c extent2, unsigned dev)
 {
@@ -449,10 +452,6 @@ static int check_extent_checksum(struct btree_trans *trans,
 	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(extent);
 	const union bch_extent_entry *entry;
 	struct extent_ptr_decoded p;
-	CLASS(printbuf, buf)();
-	void *data_buf = NULL;
-	struct bio *bio = NULL;
-	size_t bytes;
 	int ret = 0;
 
 	if (bkey_is_btree_ptr(extent.k))
@@ -466,27 +465,36 @@ found:
 	if (!p.crc.csum_type)
 		return false;
 
-	bytes = p.crc.compressed_size << 9;
-
 	struct bch_dev *ca = bch2_dev_get_ioref(c, dev, READ,
 				BCH_DEV_READ_REF_check_extent_checksums);
 	if (!ca)
 		return false;
 
-	data_buf = kvmalloc(bytes, GFP_KERNEL);
+	size_t bytes = p.crc.compressed_size << 9;
+	void *data_buf __free(kvfree) = kvmalloc(bytes, GFP_KERNEL);
 	if (!data_buf) {
-		ret = -ENOMEM;
-		goto err;
+		enumerated_ref_put(&ca->io_ref[READ],
+				   BCH_DEV_READ_REF_check_extent_checksums);
+		return -ENOMEM;
 	}
 
-	bio = bio_alloc(ca->disk_sb.bdev, buf_pages(data_buf, bytes), REQ_OP_READ, GFP_KERNEL);
+	struct bio *bio __free(bio_put) =
+		bio_alloc(ca->disk_sb.bdev, buf_pages(data_buf, bytes), REQ_OP_READ, GFP_KERNEL);
+
+	CLASS(printbuf, buf)(); /* before first goto */
+
 	bio->bi_iter.bi_sector = p.ptr.offset;
 	bch2_bio_map(bio, data_buf, bytes);
 	ret = submit_bio_wait(bio);
 	if (ret)
 		goto err;
 
-	prt_printf(&buf, "extents pointing to same space, but first extent checksum bad:\n");
+	struct nonce nonce = extent_nonce(extent.k->bversion, p.crc);
+	struct bch_csum csum = bch2_checksum(c, p.crc.csum_type, nonce, data_buf, bytes);
+	if (!bch2_crc_cmp(csum, p.crc.csum))
+		return false;
+
+	prt_printf(&buf, "extents pointing to same space, but first extent checksum bad - dropping:\n");
 	bch2_btree_id_to_text(&buf, btree);
 	prt_str(&buf, " ");
 	bch2_bkey_val_to_text(&buf, c, extent);
@@ -495,17 +503,10 @@ found:
 	prt_str(&buf, " ");
 	bch2_bkey_val_to_text(&buf, c, extent2);
 
-	struct nonce nonce = extent_nonce(extent.k->bversion, p.crc);
-	struct bch_csum csum = bch2_checksum(c, p.crc.csum_type, nonce, data_buf, bytes);
-	if (fsck_err_on(bch2_crc_cmp(csum, p.crc.csum),
-			trans, dup_backpointer_to_bad_csum_extent,
-			"%s", buf.buf))
+	if (fsck_err(trans, dup_backpointer_to_bad_csum_extent, "%s", buf.buf))
 		ret = drop_dev_and_update(trans, btree, extent, dev) ?: 1;
 fsck_err:
 err:
-	if (bio)
-		bio_put(bio);
-	kvfree(data_buf);
 	enumerated_ref_put(&ca->io_ref[READ],
 			   BCH_DEV_READ_REF_check_extent_checksums);
 	return ret;
@@ -591,7 +592,7 @@ static int check_bp_dup(struct btree_trans *trans,
 			return bp_missing(trans, extent, bp, other_bp.s_c);
 		}
 	} else {
-		ret = check_extent_checksum(trans,
+		ret = kill_replica_if_checksum_bad(trans,
 					    other_bp.v->btree_id, other_extent,
 					    bp->v.btree_id, extent,
 					    bp->k.p.inode);
@@ -600,7 +601,7 @@ static int check_bp_dup(struct btree_trans *trans,
 		if (ret)
 			return bp_missing(trans, extent, bp, other_bp.s_c);
 
-		ret = check_extent_checksum(trans, bp->v.btree_id, extent,
+		ret = kill_replica_if_checksum_bad(trans, bp->v.btree_id, extent,
 					    other_bp.v->btree_id, other_extent, bp->k.p.inode);
 		if (ret < 0)
 			return ret;
@@ -703,13 +704,17 @@ static int check_btree_root_to_backpointers(struct btree_trans *trans,
 	return check_extent_to_backpointers(trans, s, btree_id, b->c.level + 1, k);
 }
 
-static u64 mem_may_pin_bytes(struct bch_fs *c)
+static u64 system_totalram_bytes(void)
 {
 	struct sysinfo i;
 	si_meminfo(&i);
 
-	u64 mem_bytes = i.totalram * i.mem_unit;
-	return div_u64(mem_bytes * c->opts.fsck_memory_usage_percent, 100);
+	return i.totalram * i.mem_unit;
+}
+
+static u64 mem_may_pin_bytes(struct bch_fs *c)
+{
+	return div_u64(system_totalram_bytes() * c->opts.fsck_memory_usage_percent, 100);
 }
 
 static size_t btree_nodes_fit_in_ram(struct bch_fs *c)
@@ -1214,13 +1219,9 @@ int bch2_check_bucket_backpointer_mismatch(struct btree_trans *trans,
 static int check_one_backpointer(struct btree_trans *trans,
 				 struct bbpos start,
 				 struct bbpos end,
-				 struct bkey_s_c bp_k,
+				 struct bkey_s_c_backpointer bp,
 				 struct wb_maybe_flush *last_flushed)
 {
-	if (bp_k.k->type != KEY_TYPE_backpointer)
-		return 0;
-
-	struct bkey_s_c_backpointer bp = bkey_s_c_to_backpointer(bp_k);
 	struct bbpos pos = bp_to_bbpos(*bp.v);
 
 	if (bbpos_cmp(pos, start) < 0 ||
@@ -1241,12 +1242,11 @@ static int check_bucket_backpointers_to_extents(struct btree_trans *trans,
 {
 	u32 restart_count = trans->restart_count;
 
-	int ret = for_each_btree_key_max(trans, iter, BTREE_ID_backpointers,
-				      bucket_pos_to_bp_start(ca, bucket),
-				      bucket_pos_to_bp_end(ca, bucket),
-				      0, k,
-		check_one_backpointer(trans, BBPOS_MIN, BBPOS_MAX, k, last_flushed)
-	);
+	int ret = backpointer_scan_for_each(trans, iter,
+			bucket_pos_to_bp_start(ca, bucket),
+			bucket_pos_to_bp_end(ca, bucket),
+			last_flushed, NULL, bp,
+		check_one_backpointer(trans, BBPOS_MIN, BBPOS_MAX, bp, last_flushed));
 
 	return ret ?:
 		bch2_btree_write_buffer_flush_sync(trans) ?: /* make sure bad backpointers that were deleted are visible */
@@ -1264,11 +1264,9 @@ static int bch2_check_backpointers_to_extents_pass(struct btree_trans *trans,
 	bch2_progress_init(&progress, "backpointers_to_extents", trans->c,
 			   BIT_ULL(BTREE_ID_backpointers), 0);
 
-	return for_each_btree_key(trans, iter, BTREE_ID_backpointers,
-				     POS_MIN, BTREE_ITER_prefetch, k, ({
-			bch2_progress_update_iter(trans, &progress, &iter);
-			check_one_backpointer(trans, start, end, k, &last_flushed);
-	}));
+	return backpointer_scan_for_each(trans, iter, POS_MIN, POS_MAX,
+			&last_flushed, &progress, bp,
+		check_one_backpointer(trans, start, end, bp, &last_flushed));
 }
 
 int bch2_check_backpointers_to_extents(struct bch_fs *c)
@@ -1314,6 +1312,76 @@ int bch2_check_backpointers_to_extents(struct bch_fs *c)
 	return ret;
 }
 
+static int bkey_i_backpointer_cmp(const void *_l, const void *_r)
+{
+	const struct bkey_i_backpointer *l = _l;
+	const struct bkey_i_backpointer *r = _r;
+	struct bbpos l_pos = BBPOS(l->v.btree_id, l->v.pos);
+	struct bbpos r_pos = BBPOS(r->v.btree_id, r->v.pos);
+
+	/* Sort in reverse order, we'll be iterating in reverse order */
+	return -bbpos_cmp(l_pos, r_pos);
+}
+
+struct bkey_s_c_backpointer bch2_bp_scan_iter_peek(struct btree_trans *trans,
+						   struct bp_scan_iter *iter, struct bpos end,
+						   struct wb_maybe_flush *last_flushed)
+{
+	if (iter->nr_flushes != last_flushed->nr_flushes) {
+		if (iter->bps.nr) {
+			struct bkey_i_backpointer *prev = &darray_last(iter->bps);
+
+			CLASS(btree_iter, bp_iter)(trans, BTREE_ID_backpointers, prev->k.p, 0);
+			struct bkey_s_c k;
+			int ret = bkey_err(k = bch2_btree_iter_peek_slot(&bp_iter));
+			if (bkey_err(k))
+				return (struct bkey_s_c_backpointer) { .k = ERR_PTR(ret) };
+
+			if (k.k->type == KEY_TYPE_backpointer)
+				bkey_reassemble(&prev->k_i, k);
+			else
+				--iter->bps.nr;
+		}
+
+		iter->nr_flushes = last_flushed->nr_flushes;
+	}
+
+	if (!iter->bps.nr) {
+		size_t limit = (system_totalram_bytes() / 16) /
+			sizeof(struct bkey_i_backpointer);
+
+		u32 restart_count = trans->restart_count;
+
+		int ret = for_each_btree_key_max(trans, bp_iter, BTREE_ID_backpointers,
+					   iter->pos, end, BTREE_ITER_prefetch, k, ({
+			if (k.k->type != KEY_TYPE_backpointer)
+				continue;
+
+			struct bkey_i_backpointer bp;
+			bkey_reassemble(&bp.k_i, k);
+			if (iter->bps.nr > limit ||
+			    darray_push_gfp(&iter->bps, bp, GFP_KERNEL|__GFP_NOWARN))
+				break;
+
+			iter->pos = bpos_nosnap_successor(k.k->p);
+			(iter->progress
+			 ? bch2_progress_update_iter(trans, iter->progress, &bp_iter)
+			 : 0);
+		}));
+		trans->restart_count = restart_count;
+
+		if (ret)
+			return ((struct bkey_s_c_backpointer) { .k = ERR_PTR(ret) });
+
+		if (!iter->bps.nr)
+			return (struct bkey_s_c_backpointer) {};
+
+		darray_sort(iter->bps, bkey_i_backpointer_cmp);
+	}
+
+	return backpointer_i_to_s_c(&darray_last(iter->bps));
+}
+
 static int bch2_bucket_bitmap_set(struct bch_dev *ca, struct bucket_bitmap *b, u64 bit)
 {
 	scoped_guard(mutex, &b->lock) {
