@@ -13,7 +13,7 @@
 
 #include "data/copygc.h"
 #include "data/ec.h"
-#include "data/rebalance.h"
+#include "data/reconcile.h"
 
 #include "fs/check.h"
 #include "fs/inode.h"
@@ -87,6 +87,7 @@ static int bch2_sb_recovery_passes_validate(struct bch_sb *sb, struct bch_sb_fie
 }
 
 static void bch2_sb_recovery_passes_to_text(struct printbuf *out,
+					    struct bch_fs *c,
 					    struct bch_sb *sb,
 					    struct bch_sb_field *f)
 {
@@ -97,7 +98,7 @@ static void bch2_sb_recovery_passes_to_text(struct printbuf *out,
 	if (out->nr_tabstops < 1)
 		printbuf_tabstop_push(out, 32);
 	if (out->nr_tabstops < 2)
-		printbuf_tabstop_push(out, 16);
+		printbuf_tabstop_push(out, 32);
 
 	prt_printf(out, "Pass\tLast run\tLast runtime\n");
 
@@ -112,7 +113,7 @@ static void bch2_sb_recovery_passes_to_text(struct printbuf *out,
 		bch2_prt_datetime(out, le64_to_cpu(i->last_run));
 		prt_tab(out);
 
-		bch2_pr_time_units(out, le32_to_cpu(i->last_runtime) * NSEC_PER_SEC);
+		bch2_pr_time_units(out, (u64) le32_to_cpu(i->last_runtime) * NSEC_PER_SEC);
 
 		if (BCH_RECOVERY_PASS_NO_RATELIMIT(i))
 			prt_str(out, " (no ratelimit)");
@@ -148,11 +149,15 @@ static void bch2_sb_recovery_pass_complete(struct bch_fs *c,
 					   enum bch_recovery_pass pass,
 					   s64 start_time)
 {
+	guard(memalloc_flags)(PF_MEMALLOC_NOFS);
 	guard(mutex)(&c->sb_lock);
 	struct bch_sb_field_ext *ext = bch2_sb_field_get(c->disk_sb.sb, ext);
 	__clear_bit_le64(bch2_recovery_pass_to_stable(pass),
 			 ext->recovery_passes_required);
 
+	if (bch2_is_zero(ext->recovery_passes_required, sizeof(ext->recovery_passes_required)))
+		memset(ext->errors_silent, 0, sizeof(ext->errors_silent));
+
 	struct recovery_pass_entry *e = bch2_sb_recovery_pass_entry(c, pass);
 	if (e) {
 		s64 end_time	= ktime_get_real_seconds();
@@ -167,6 +172,7 @@ static void bch2_sb_recovery_pass_complete(struct bch_fs *c,
 void bch2_recovery_pass_set_no_ratelimit(struct bch_fs *c,
 					 enum bch_recovery_pass pass)
 {
+	guard(memalloc_flags)(PF_MEMALLOC_NOFS);
 	guard(mutex)(&c->sb_lock);
 
 	struct recovery_pass_entry *e = bch2_sb_recovery_pass_entry(c, pass);
@@ -176,31 +182,44 @@ void bch2_recovery_pass_set_no_ratelimit(struct bch_fs *c,
 	}
 }
 
-static bool bch2_recovery_pass_want_ratelimit(struct bch_fs *c, enum bch_recovery_pass pass)
+static bool bch2_recovery_pass_entry_get_locked(struct bch_fs *c, enum bch_recovery_pass pass,
+						struct recovery_pass_entry *e)
 {
-	enum bch_recovery_pass_stable stable = bch2_recovery_pass_to_stable(pass);
-	bool ret = false;
-
 	lockdep_assert_held(&c->sb_lock);
 
 	struct bch_sb_field_recovery_passes *r =
 		bch2_sb_field_get(c->disk_sb.sb, recovery_passes);
 
-	if (stable < recovery_passes_nr_entries(r)) {
-		struct recovery_pass_entry *i = r->start + stable;
+	enum bch_recovery_pass_stable stable = bch2_recovery_pass_to_stable(pass);
+	bool found = stable < recovery_passes_nr_entries(r);
+	if (found)
+		*e = r->start[stable];
 
-		/*
-		 * Ratelimit if the last runtime was more than 1% of the time
-		 * since we last ran
-		 */
-		ret = (u64) le32_to_cpu(i->last_runtime) * 100 >
-			ktime_get_real_seconds() - le64_to_cpu(i->last_run);
+	return found;
+}
 
-		if (BCH_RECOVERY_PASS_NO_RATELIMIT(i))
-			ret = false;
-	}
+static bool bch2_recovery_pass_want_ratelimit_locked(struct bch_fs *c, enum bch_recovery_pass pass,
+						     unsigned runtime_fraction)
+{
+	struct recovery_pass_entry e;
+	if (!bch2_recovery_pass_entry_get_locked(c, pass, &e))
+		return false;
 
-	return ret;
+	/*
+	 * Ratelimit if the last runtime was more than 1% of the time
+	 * since we last ran
+	 */
+	return !BCH_RECOVERY_PASS_NO_RATELIMIT(&e) &&
+		(u64) le32_to_cpu(e.last_runtime) * runtime_fraction >
+		ktime_get_real_seconds() - le64_to_cpu(e.last_run);
+}
+
+bool bch2_recovery_pass_want_ratelimit(struct bch_fs *c, enum bch_recovery_pass pass,
+				       unsigned runtime_fraction)
+{
+	guard(memalloc_flags)(PF_MEMALLOC_NOFS);
+	guard(mutex)(&c->sb_lock);
+	return bch2_recovery_pass_want_ratelimit_locked(c, pass, runtime_fraction);
 }
 
 const struct bch_sb_field_ops bch_sb_field_ops_recovery_passes = {
@@ -214,30 +233,6 @@ static int bch2_recovery_pass_empty(struct bch_fs *c)
 	return 0;
 }
 
-static int bch2_set_may_go_rw(struct bch_fs *c)
-{
-	struct journal_keys *keys = &c->journal_keys;
-
-	/*
-	 * After we go RW, the journal keys buffer can't be modified (except for
-	 * setting journal_key->overwritten: it will be accessed by multiple
-	 * threads
-	 */
-	move_gap(keys, keys->nr);
-
-	set_bit(BCH_FS_may_go_rw, &c->flags);
-
-	if (go_rw_in_recovery(c)) {
-		if (c->sb.features & BIT_ULL(BCH_FEATURE_no_alloc_info)) {
-			bch_info(c, "mounting a filesystem with no alloc info read-write; will recreate");
-			bch2_reconstruct_alloc(c);
-		}
-
-		return bch2_fs_read_write_early(c);
-	}
-	return 0;
-}
-
 /*
  * Make sure root inode is readable while we're still in recovery and can rewind
  * for repair:
@@ -254,24 +249,30 @@ static int bch2_lookup_root_inode(struct bch_fs *c)
 		bch2_inode_find_by_inum_trans(trans, inum, &inode_u));
 }
 
-struct recovery_pass_fn {
+struct recovery_pass {
 	int		(*fn)(struct bch_fs *);
 	const char	*name;
 	unsigned	when;
+	u64		depends;
 };
 
-static struct recovery_pass_fn recovery_pass_fns[] = {
-#define x(_fn, _id, _when)	{ .fn = bch2_##_fn, .name = #_fn, .when = _when },
+static const struct recovery_pass recovery_passes[] = {
+#define x(_fn, _id, _when, _depends)	{	\
+	.fn		= bch2_##_fn,		\
+	.name		= #_fn,			\
+	.when		= _when,		\
+	.depends	= _depends,		\
+},
 	BCH_RECOVERY_PASSES()
 #undef x
 };
 
-static u64 bch2_recovery_passes_match(unsigned flags)
+u64 bch2_recovery_passes_match(unsigned flags)
 {
 	u64 ret = 0;
 
-	for (unsigned i = 0; i < ARRAY_SIZE(recovery_pass_fns); i++)
-		if (recovery_pass_fns[i].when & flags)
+	for (unsigned i = 0; i < ARRAY_SIZE(recovery_passes); i++)
+		if (recovery_passes[i].when & flags)
 			ret |= BIT_ULL(i);
 	return ret;
 }
@@ -281,20 +282,41 @@ u64 bch2_fsck_recovery_passes(void)
 	return bch2_recovery_passes_match(PASS_FSCK);
 }
 
-static void bch2_run_async_recovery_passes(struct bch_fs *c)
+/* Set of all passes that depend on @pass, transitively */
+static u64 pass_dependents(enum bch_recovery_pass pass)
 {
-	if (!down_trylock(&c->recovery.run_lock))
-		return;
+	u64 passes = BIT_ULL(pass);
+	bool found;
 
-	if (!enumerated_ref_tryget(&c->writes, BCH_WRITE_REF_async_recovery_passes))
-		goto unlock;
+	do {
+		found = false;
+		for (unsigned i = 0; i < BCH_RECOVERY_PASS_NR; i++)
+			if (!(passes & BIT_ULL(i)) &&
+			    (passes & recovery_passes[i].depends)) {
+				passes |= BIT_ULL(i);
+				found = true;
+			}
+	} while (found);
 
-	if (queue_work(system_long_wq, &c->recovery.work))
-		return;
+	return passes;
+}
 
-	enumerated_ref_put(&c->writes, BCH_WRITE_REF_async_recovery_passes);
-unlock:
-	up(&c->recovery.run_lock);
+/* Returns true if a given pass and all scheduled dependents can run online */
+static bool recovery_pass_should_defer(enum bch_recovery_pass pass,
+				       u64 passes)
+{
+	passes &= pass_dependents(pass);
+	passes |= BIT_ULL(pass);
+	return passes == (passes & bch2_recovery_passes_match(PASS_ONLINE));
+}
+
+static bool recovery_pass_needs_rewind(struct bch_fs *c,
+				       enum bch_recovery_pass pass)
+{
+	struct bch_fs_recovery *r = &c->recovery;
+	return  test_bit(BCH_FS_in_recovery, &c->flags) &&
+		r->current_pass > pass &&
+		!(r->passes_complete & BIT_ULL(pass));
 }
 
 static bool recovery_pass_needs_set(struct bch_fs *c,
@@ -311,7 +333,7 @@ static bool recovery_pass_needs_set(struct bch_fs *c,
 		*flags |= RUN_RECOVERY_PASS_nopersistent;
 
 	if ((*flags & RUN_RECOVERY_PASS_ratelimit) &&
-	    !bch2_recovery_pass_want_ratelimit(c, pass))
+	    !bch2_recovery_pass_want_ratelimit_locked(c, pass, 100))
 		*flags &= ~RUN_RECOVERY_PASS_ratelimit;
 
 	/*
@@ -325,23 +347,18 @@ static bool recovery_pass_needs_set(struct bch_fs *c,
 	 */
 	bool in_recovery = test_bit(BCH_FS_in_recovery, &c->flags);
 	bool persistent = !in_recovery || !(*flags & RUN_RECOVERY_PASS_nopersistent);
-	bool rewind = in_recovery &&
-		r->curr_pass > pass &&
-		!(r->passes_complete & BIT_ULL(pass));
+	u64 already_running = persistent
+		? c->sb.recovery_passes_required
+		: r->current_passes;
 
-	if (persistent
-	    ? !(c->sb.recovery_passes_required & BIT_ULL(pass))
-	    : !((r->passes_to_run|r->passes_complete) & BIT_ULL(pass)))
+	if (!(already_running & BIT_ULL(pass)))
 		return true;
 
 	if (!(*flags & RUN_RECOVERY_PASS_ratelimit) &&
 	    (r->passes_ratelimiting & BIT_ULL(pass)))
 		return true;
 
-	if (rewind)
-		return true;
-
-	return false;
+	return recovery_pass_needs_rewind(c, pass);
 }
 
 /*
@@ -354,7 +371,6 @@ int __bch2_run_explicit_recovery_pass(struct bch_fs *c,
 				      bool *write_sb)
 {
 	struct bch_fs_recovery *r = &c->recovery;
-	int ret = 0;
 
 	lockdep_assert_held(&c->sb_lock);
 
@@ -365,20 +381,21 @@ int __bch2_run_explicit_recovery_pass(struct bch_fs *c,
 	if (!recovery_pass_needs_set(c, pass, &flags))
 		return 0;
 
+	out->suppress = false;
+
 	bool in_recovery = test_bit(BCH_FS_in_recovery, &c->flags);
-	bool rewind = in_recovery &&
-		r->curr_pass > pass &&
-		!(r->passes_complete & BIT_ULL(pass));
 	bool ratelimit = flags & RUN_RECOVERY_PASS_ratelimit;
 
-	if (!(flags & RUN_RECOVERY_PASS_nopersistent)) {
+	if (flags & RUN_RECOVERY_PASS_nopersistent) {
+		r->scheduled_passes_ephemeral |= BIT_ULL(pass);
+	} else {
 		struct bch_sb_field_ext *ext = bch2_sb_field_get(c->disk_sb.sb, ext);
 		*write_sb |= !__test_and_set_bit_le64(bch2_recovery_pass_to_stable(pass),
 						     ext->recovery_passes_required);
 	}
 
 	if (pass < BCH_RECOVERY_PASS_set_may_go_rw &&
-	    (!in_recovery || r->curr_pass >= BCH_RECOVERY_PASS_set_may_go_rw)) {
+	    test_bit(BCH_FS_may_go_rw, &c->flags)) {
 		prt_printf(out, "need recovery pass %s (%u), but already rw\n",
 			   bch2_recovery_passes[pass], pass);
 		return bch_err_throw(c, cannot_rewind_recovery);
@@ -389,30 +406,33 @@ int __bch2_run_explicit_recovery_pass(struct bch_fs *c,
 	else
 		r->passes_ratelimiting &= ~BIT_ULL(pass);
 
-	if (in_recovery && !ratelimit) {
+	if (in_recovery && !ratelimit && !recovery_pass_should_defer(pass, r->current_passes)) {
+		bool rewind = recovery_pass_needs_rewind(c, pass);
+
 		prt_printf(out, "running recovery pass %s (%u), currently at %s (%u)%s\n",
 			   bch2_recovery_passes[pass], pass,
-			   bch2_recovery_passes[r->curr_pass], r->curr_pass,
+			   bch2_recovery_passes[r->current_pass], r->current_pass,
 			   rewind ? " - rewinding" : "");
 
-		r->passes_to_run |= BIT_ULL(pass);
+		r->current_passes |= BIT_ULL(pass);
 
 		if (rewind) {
-			r->next_pass = pass;
-			r->passes_complete &= (1ULL << pass) >> 1;
-			ret = bch_err_throw(c, restart_recovery);
+			r->rewound_to = r->rewound_to
+				? min(r->rewound_to, pass)
+				: pass;
+			return bch_err_throw(c, restart_recovery);
 		}
 	} else {
 		prt_printf(out, "scheduling recovery pass %s (%u)%s\n",
 			   bch2_recovery_passes[pass], pass,
 			   ratelimit ? " - ratelimiting" : "");
 
-		struct recovery_pass_fn *p = recovery_pass_fns + pass;
-		if (p->when & PASS_ONLINE)
+		const struct recovery_pass *p = recovery_passes + pass;
+		if (!ratelimit && (p->when & PASS_ONLINE))
 			bch2_run_async_recovery_passes(c);
 	}
 
-	return ret;
+	return 0;
 }
 
 int bch2_run_explicit_recovery_pass(struct bch_fs *c,
@@ -428,6 +448,7 @@ int bch2_run_explicit_recovery_pass(struct bch_fs *c,
 	    !recovery_pass_needs_set(c, pass, &flags))
 		return 0;
 
+	guard(memalloc_flags)(PF_MEMALLOC_NOFS);
 	guard(mutex)(&c->sb_lock);
 	bool write_sb = false;
 	int ret = __bch2_run_explicit_recovery_pass(c, out, pass, flags, &write_sb);
@@ -449,9 +470,10 @@ int bch2_require_recovery_pass(struct bch_fs *c,
 	    c->recovery.passes_complete & BIT_ULL(pass))
 		return 0;
 
+	guard(memalloc_flags)(PF_MEMALLOC_NOFS);
 	guard(mutex)(&c->sb_lock);
 
-	if (bch2_recovery_pass_want_ratelimit(c, pass))
+	if (bch2_recovery_pass_want_ratelimit_locked(c, pass, 100))
 		return 0;
 
 	enum bch_run_recovery_pass_flags flags = 0;
@@ -464,30 +486,10 @@ int bch2_require_recovery_pass(struct bch_fs *c,
 	return ret;
 }
 
-int bch2_run_print_explicit_recovery_pass(struct bch_fs *c, enum bch_recovery_pass pass)
-{
-	enum bch_run_recovery_pass_flags flags = 0;
-
-	if (!recovery_pass_needs_set(c, pass, &flags))
-		return 0;
-
-	CLASS(printbuf, buf)();
-	bch2_log_msg_start(c, &buf);
-
-	guard(mutex)(&c->sb_lock);
-	bool write_sb = false;
-	int ret = __bch2_run_explicit_recovery_pass(c, &buf, pass,
-						RUN_RECOVERY_PASS_nopersistent,
-						&write_sb);
-
-	bch2_print_str(c, KERN_NOTICE, buf.buf);
-	return ret;
-}
-
 static int bch2_run_recovery_pass(struct bch_fs *c, enum bch_recovery_pass pass)
 {
 	struct bch_fs_recovery *r = &c->recovery;
-	struct recovery_pass_fn *p = recovery_pass_fns + pass;
+	const struct recovery_pass *p = recovery_passes + pass;
 
 	if (!(p->when & PASS_SILENT))
 		bch2_print(c, KERN_INFO bch2_log_msg(c, "%s..."),
@@ -495,37 +497,30 @@ static int bch2_run_recovery_pass(struct bch_fs *c, enum bch_recovery_pass pass)
 
 	s64 start_time = ktime_get_real_seconds();
 	int ret = p->fn(c);
-
-	r->passes_to_run &= ~BIT_ULL(pass);
-
 	if (ret) {
 		bch_err(c, "%s(): error %s", p->name, bch2_err_str(ret));
 		r->passes_failing |= BIT_ULL(pass);
 		return ret;
 	}
 
+	if (!(p->when & PASS_SILENT))
+		bch2_print(c, KERN_CONT " done (%lli seconds)\n",
+			   ktime_get_real_seconds() - start_time);
 	r->passes_failing = 0;
 
 	if (!test_bit(BCH_FS_error, &c->flags))
 		bch2_sb_recovery_pass_complete(c, pass, start_time);
 
-	if (!(p->when & PASS_SILENT))
-		bch2_print(c, KERN_CONT " done\n");
-
 	return 0;
 }
 
-static int __bch2_run_recovery_passes(struct bch_fs *c, u64 orig_passes_to_run,
-				      bool online)
+int bch2_run_recovery_passes(struct bch_fs *c, u64 orig_passes_to_run, bool failfast)
 {
 	struct bch_fs_recovery *r = &c->recovery;
 	int ret = 0;
 
 	spin_lock_irq(&r->lock);
 
-	if (online)
-		orig_passes_to_run &= bch2_recovery_passes_match(PASS_ONLINE);
-
 	if (c->sb.features & BIT_ULL(BCH_FEATURE_no_alloc_info))
 		orig_passes_to_run &= ~bch2_recovery_passes_match(PASS_ALLOC);
 
@@ -539,14 +534,15 @@ static int __bch2_run_recovery_passes(struct bch_fs *c, u64 orig_passes_to_run,
 
 	orig_passes_to_run &= ~r->passes_failing;
 
-	r->passes_to_run = orig_passes_to_run;
+	r->current_passes = orig_passes_to_run;
 
-	while (r->passes_to_run) {
-		unsigned prev_done = r->pass_done;
-		unsigned pass = __ffs64(r->passes_to_run);
-		r->curr_pass = pass;
-		r->next_pass = r->curr_pass + 1;
-		r->passes_to_run &= ~BIT_ULL(pass);
+	enum bch_recovery_pass prev = 0;
+	while (r->current_passes) {
+		unsigned pass = __ffs64(r->current_passes);
+
+		r->current_pass			= pass;
+		r->current_passes		&= ~BIT_ULL(pass);
+		r->scheduled_passes_ephemeral	&= ~BIT_ULL(pass);
 
 		spin_unlock_irq(&r->lock);
 
@@ -555,9 +551,11 @@ static int __bch2_run_recovery_passes(struct bch_fs *c, u64 orig_passes_to_run,
 
 		spin_lock_irq(&r->lock);
 
-		if (r->next_pass < r->curr_pass) {
-			/* Rewind: */
-			r->passes_to_run |= orig_passes_to_run & (~0ULL << r->next_pass);
+		if (r->rewound_to) {
+			r->rewound_from	= max(r->rewound_from, pass);
+			/* Restore r->current_passses up to and including r->rewound_to */
+			r->current_passes |= orig_passes_to_run & (~0ULL << r->rewound_to);
+			r->rewound_to = 0;
 		} else if (!ret2) {
 			r->pass_done = max(r->pass_done, pass);
 			r->passes_complete |= BIT_ULL(pass);
@@ -565,17 +563,19 @@ static int __bch2_run_recovery_passes(struct bch_fs *c, u64 orig_passes_to_run,
 			ret = ret2;
 		}
 
-		if (ret && !online)
+		if (ret && failfast)
 			break;
 
-		if (prev_done <= BCH_RECOVERY_PASS_check_snapshots &&
-		    r->pass_done > BCH_RECOVERY_PASS_check_snapshots) {
+		if (prev <= BCH_RECOVERY_PASS_check_snapshots &&
+		    pass > BCH_RECOVERY_PASS_check_snapshots) {
 			bch2_copygc_wakeup(c);
-			bch2_rebalance_wakeup(c);
+			bch2_reconcile_wakeup(c);
 		}
+
+		prev = pass;
 	}
 
-	clear_bit(BCH_FS_in_recovery, &c->flags);
+	r->current_pass = 0;
 	spin_unlock_irq(&r->lock);
 
 	return ret;
@@ -586,21 +586,36 @@ static void bch2_async_recovery_passes_work(struct work_struct *work)
 	struct bch_fs *c = container_of(work, struct bch_fs, recovery.work);
 	struct bch_fs_recovery *r = &c->recovery;
 
-	__bch2_run_recovery_passes(c,
-		c->sb.recovery_passes_required & ~r->passes_ratelimiting,
-		true);
+	if (mutex_trylock(&r->run_lock)) {
+		bch2_run_recovery_passes(c,
+			(c->sb.recovery_passes_required |
+			 r->scheduled_passes_ephemeral) &
+			~r->passes_ratelimiting &
+			bch2_recovery_passes_match(PASS_ONLINE),
+			false);
 
-	up(&r->run_lock);
+		mutex_unlock(&r->run_lock);
+	}
 	enumerated_ref_put(&c->writes, BCH_WRITE_REF_async_recovery_passes);
 }
 
-int bch2_run_online_recovery_passes(struct bch_fs *c, u64 passes)
+void bch2_run_async_recovery_passes(struct bch_fs *c)
 {
-	return __bch2_run_recovery_passes(c, c->sb.recovery_passes_required|passes, true);
+	if (!enumerated_ref_tryget(&c->writes, BCH_WRITE_REF_async_recovery_passes))
+		return;
+
+	if (queue_work(system_long_wq, &c->recovery.work))
+		return;
+
+	enumerated_ref_put(&c->writes, BCH_WRITE_REF_async_recovery_passes);
 }
 
-int bch2_run_recovery_passes(struct bch_fs *c, enum bch_recovery_pass from)
+int bch2_run_recovery_passes_startup(struct bch_fs *c, enum bch_recovery_pass from)
 {
+	struct bch_fs_recovery *r = &c->recovery;
+
+	r->scheduled_passes_ephemeral = c->opts.recovery_passes;
+
 	u64 passes =
 		bch2_recovery_passes_match(PASS_ALWAYS) |
 		(!c->sb.clean ? bch2_recovery_passes_match(PASS_UNCLEAN) : 0) |
@@ -620,11 +635,36 @@ int bch2_run_recovery_passes(struct bch_fs *c, enum bch_recovery_pass from)
 
 	passes &= ~(BIT_ULL(from) - 1);
 
-	down(&c->recovery.run_lock);
-	int ret = __bch2_run_recovery_passes(c, passes, false);
-	up(&c->recovery.run_lock);
+	/*
+	 * Defer passes that can be run online, and don't have dependents that
+	 * can't be run online
+	 */
+	u64 defer = 0;
+	if (!c->opts.fsck)
+		for (unsigned i = 0; i < BCH_RECOVERY_PASS_NR; i++)
+			if ((passes & BIT_ULL(i)) &&
+			    !(c->opts.recovery_passes & BIT_ULL(i)) &&
+			    !(recovery_passes[i].when & PASS_NODEFER) &&
+			    recovery_pass_should_defer(i, passes)) {
+				defer |= BIT_ULL(i);
+				passes &= ~BIT_ULL(i);
+			}
 
-	return ret;
+	scoped_guard(mutex, &r->run_lock)
+		try(bch2_run_recovery_passes(c, passes, true));
+
+	clear_bit(BCH_FS_in_recovery, &c->flags);
+
+	if (defer) {
+		CLASS(bch_log_msg_level, msg)(c, LOGLEVEL_notice);
+		prt_printf(&msg.m, "Running the following recovery passes in the background:\n");
+		prt_bitflags(&msg.m, bch2_recovery_passes, defer);
+
+		r->scheduled_passes_ephemeral |= defer;
+		bch2_run_async_recovery_passes(c);
+	}
+
+	return 0;
 }
 
 static void prt_passes(struct printbuf *out, const char *msg, u64 passes)
@@ -639,24 +679,28 @@ void bch2_recovery_pass_status_to_text(struct printbuf *out, struct bch_fs *c)
 	struct bch_fs_recovery *r = &c->recovery;
 
 	printbuf_tabstop_push(out, 32);
-	prt_passes(out, "Scheduled passes", c->sb.recovery_passes_required);
-	prt_passes(out, "Scheduled online passes", c->sb.recovery_passes_required &
-		   bch2_recovery_passes_match(PASS_ONLINE));
-	prt_passes(out, "Complete passes", r->passes_complete);
-	prt_passes(out, "Failing passes", r->passes_failing);
+	prt_passes(out, "Scheduled (superblock)",	c->sb.recovery_passes_required);
+	prt_passes(out, "Scheduled (ephemeral)",	r->scheduled_passes_ephemeral);
 
-	if (r->curr_pass) {
-		prt_printf(out, "Current pass:\t%s\n", bch2_recovery_passes[r->curr_pass]);
-		prt_passes(out, "Current passes", r->passes_to_run);
+	prt_passes(out, "Completed",	r->passes_complete);
+	prt_passes(out, "Failing",	r->passes_failing);
+
+	if (r->current_pass) {
+		prt_printf(out, "Currently running:\t%s (%u)\n",
+			   bch2_recovery_passes[r->current_pass], r->current_pass);
+		prt_passes(out, "Next", r->current_passes);
+
+		if (test_bit(BCH_FS_in_recovery, &c->flags) && r->rewound_from)
+			prt_printf(out, "Rewound from:\t%s (%u)\n",
+				   bch2_recovery_passes[r->rewound_from],
+				   r->rewound_from);
 	}
-
-	prt_printf(out, "Pass done:\t%s\n", bch2_recovery_passes[r->pass_done]);
 }
 
 void bch2_fs_recovery_passes_init(struct bch_fs *c)
 {
 	spin_lock_init(&c->recovery.lock);
-	sema_init(&c->recovery.run_lock, 1);
+	mutex_init(&c->recovery.run_lock);
 
 	INIT_WORK(&c->recovery.work, bch2_async_recovery_passes_work);
 }
